-- Puffin: A Tojaso FowlPlay Project
-- Adapter component for creating and updating lists of auras or cooldowns.
-- Each list has an associated icon group view which gets configured during initialization.
-- Lists are filtered, sorted and annotated in response to changes in models for unit auras or cooldowns.
-- Icon group views are updated when associated spell lists are changed.

local Adapter_IconGroups = {}
local EF = Puffin.ErrorFormat
local DB, UN, IG, DP, LSM
local active_groups = {}
local configuration_changed = true
local units_aura_tracked = {}
local units_aura_changed = {}
local cooldowns_tracked = false
local cooldowns_changed = false

-- Track aura changes to units associated with icon groups
local function UnitChanged(e, unit, why) if why == "auras" then units_aura_changed[unit] = true end end

-- Track cooldown changes to support icon groups that display spells on cooldown
local function CooldownsChanged(e) cooldowns_changed = true end

-- Check an attribute to see if it refers to special media
local function CheckMedia(v)
	if type(v) == "table" and v[1] == "#LSM" then v = LSM:Fetch(v[2], v[3]) end -- check for LibSharedMedia references
	return v
end

-- Check status for a group to see if it should be hidden
local function GroupStatus(gp)
	return false
end

-- Apply filters to a spell from a given source to see if it should be displayed in the group
local function GroupFilter(gp, src, spell)
	return true
end

-- Apply sorters to the spells in a group, use simple bubble sort since spell lists are never long
-- Available sorters include: AtoZ, ZtoA, TimeUp, TimeDown, DurationUp, DurationDown, StartUp, StartDown
local function GroupSorter(gp, spells)
	local sorters = gp.sorters
	local count = #spells
	if sorters and count > 1 then
		for i = 1, count - 1 do
			for j = i + 1, count do
				local k, kmax, flip, done, s1, s2 = 0, #sorters, false, false, spells[i], spells[j]
				while k < kmax and not done do
					k = k + 1
					local s = sorters[k]
					local v1, v2
					if s == "AtoZ" then v1 = s1.name; v2 = s2.name
					elseif s == "ZtoA" then v2 = s1.name; v1 = s2.name
					elseif s == "TimeUp" then v1 = s1.expire; v2 = s2.expire
					elseif s == "TimeDown" then v2 = s1.expire; v1 = s2.expire
					elseif s == "DurationUp" then v1 = s1.duration; v2 = s2.duration
					elseif s == "DurationDown" then v2 = s1.duration; v1 = s2.duration
					elseif s == "StartUp" then v1 = s1.expire - s1.duration; v2 = s2.expire - s2.duration
					elseif s == "StartDown" then v2 = s1.expire - s1.duration; v1 = s2.expire - s2.duration
					end
					if v1 ~= v2 then flip = (v2 ~= 0) and (v1 > v2); done = true end
				end
				if flip then spells[i] = s2; spells[j] = s1 end
			end
		end
	end
end

function Adapter_IconGroups.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("Adapter_IconGroups", "Data component not found"))
	UN = Puffin.Component("Model_Units")
	assert(UN, EF("Adapter_IconGroups", "Model_Units component not found"))
	IG = Puffin.Component("View_IconGroup")
	assert(IG, EF("Adapter_IconGroups", "View_IconGroup component not found"))
	LSM = LibStub("LibSharedMedia-3.0")
	assert(LSM, EF("Adapter_Configuration", "LibSharedMedia not found"))
end

-- On first update, configure the icon groups, afterwards just update each as needed
function Adapter_IconGroups.OnUpdate()
	local layout = DB.GetRecord("icon_groups", "layout")
	local groups = DB.GetRecord("icon_groups", "groups")
	if not layout or not groups then return end

	if configuration_changed then -- can not configure during combat lockdown
		for name, gp in pairs(groups) do -- allocate and configure each group and associated view
			local u = IG.Create() -- create the icon group view
			u.defaults = layout.defaults; u.widgets = layout.widgets
			if gp.anchor then
				u.offsetX = gp.anchor.offsetX; u.offsetY = gp.anchor.offsetY
				u.anchorPoint = gp.anchor.anchorPoint; u.relativePoint = gp.anchor.relativePoint
			else u.offsetX = 0; u.offsetY = 0; u.anchorPoint = "CENTER"; u.relativePoint = "CENTER" end
			IG.Configure(u)
			active_groups[name] = { view = u, group = gp, hidden = false, spells = {} }
			if gp.sources then
				for _, src in ipairs(gp.sources) do
					local unit = src.unit
					if unit and (src.data == "buffs" or src.data == "debuffs") then
						if not units_aura_tracked[unit] then
							units_aura_tracked[unit] = true; units_aura_changed[unit] = true
							Puffin.RegisterCallback(UN.callbacks[unit], UnitChanged)
						end
					elseif src.data == "cooldowns" then
						if not cooldowns_tracked then
							cooldowns_tracked = true; cooldowns_changed = true
							Puffin.RegisterCallback("Cooldowns", CooldownsChanged)
						end
					end
				end
			end
		end
		configuration_changed = false -- don't process icon groups first time through
		return
	end
	
	for name, group in pairs(active_groups) do -- update all icon group views
		local gp = group.group
		if gp and gp.sources then
			local update_group, hide_group = false, GroupStatus(gp) -- check status to see if group is hidden
			if not hide_group or gp.hidden then
				for _, src in ipairs(gp.sources) do -- iterate through sources and see if need to update due to changes in model
					if src.unit and (src.data == "buffs" or src.data == "debuffs") then -- process aura-related sources
						if units_aura_changed[src.unit] then update_group = true end
					elseif src.data == "cooldowns" then
						if cooldowns_changed then update_group = true end
					end
				end
				if hide_group or update_group then
					for _, s in ipairs(group.spells) do Puffin.ReleaseTable(s) end -- release previous spell list contents
					table.wipe(group.spells) -- list of spells to be displayed in the icon group
					local count = 0
					if not hide_group then
						for _, src in ipairs(gp.sources) do -- iterate through sources, filter each spell before adding to the list
							if src.unit and (src.data == "buffs" or src.data == "debuffs") then -- process aura-related sources
								local t = DB.GetRecord("units", src.unit)
								if t and t.auras then
									local ap = t.auras[src.data]
									for i = 1, ap.count do
										local a = ap[i]
										if GroupFilter(gp, src, a) then -- apply filters to this spell to make sure it should be included in the list
											local s = Puffin.AllocateTable()
											s.name = a.name; s.duration = a.duration; s.expire = a.expire
											s.count = a.count; s.icon = a.icon
											-- need to add tooltip info
											count = count + 1; group.spells[count] = s
										end
									end
								end
							elseif src.data == "cooldowns" then
								for cname, cd in DB.GetRecords("cooldowns") do
									if GroupFilter(gp, src, cd) then
										local s = Puffin.AllocateTable()
										s.name = cname; s.duration = cd.duration; s.expire = cd.start + cd.duration
										s.count = cd.count; s.icon = cd.icon
										-- need to add tooltip info
										count = count + 1; group.spells[count] = s
									end
								end
							end
						end
						GroupSorter(gp, group.spells) -- sort the spell list
					end
					-- print("spells", name, count, GetTime())
					-- for i, s in ipairs(group.spells) do print(i, s.name, s.duration, s.expire, s.count) end
					-- IG.Update(group.view)
				end
			end
		end
	end
	
	table.wipe(units_aura_changed); cooldowns_changed = false
end

Puffin.RegisterComponent("Adapter_IconGroups", Adapter_IconGroups, "ADAPTER", 3)
