-- Puffin: A Tojaso FowlPlay Project
-- Model component for tracking cooldown information for player and pet
-- Each cooldown is stored in the database with a record identified by the cooldown name
-- Record attributes: cdType, id, icon, start, duration, timeLeft, ttType, ttArg

local Model_Cooldowns = {}
local DB = nil
local recycled = {}
local startGCD, durationGCD = nil -- detect global cooldowns
local cooldown_changed = true -- indicates there has been an event that requires updating cooldowns
local cds = {} -- temporary table for record names
local cooldown = {} -- temporary table for entering a cooldown
local iconGCD = nil -- icon to use for GCD
local iconPotion = nil -- icon for shared potions cooldown
local iconElixir = nil -- icon for shared elixirs cooldown
local runeSlots = {} -- cache of info about death knight runes
local lockouts = {} -- schools of magic that we are currently locked out of
local lockstarts = {} -- start times for current school lockouts

local slots = { "HeadSlot", "NeckSlot", "BackSlot", "ShoulderSlot", "ChestSlot", "ShirtSlot", "TabardSlot", "WristSlot", "HandsSlot",
	"WaistSlot", "LegsSlot", "FeetSlot", "Finger0Slot", "Finger1Slot", "Trinket0Slot", "Trinket1Slot", "MainHandSlot", "SecondaryHandSlot" }

local function CooldownEvent(e) cooldown_changed = true end

-- Function called to detect global cooldowns
local function CheckGCD(event, unit, spell)
	if unit == "player" then
		local start, duration = GetSpellCooldown(spell)
		if start and duration and (duration > 0) and (duration <= 1.5) then startGCD = start; durationGCD = duration; cooldown_changed = true end
	end
end

-- Initialize by creating a database for cooldowns (assumes records will not be redirected)
function Model_Cooldowns.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, "Puffin.Model_Cooldowns: invalid reference to data management component")
	DB.AllocateDatabase("cooldowns", false)
	
	Puffin.RegisterCallback("SPELL_UPDATE_COOLDOWN", CooldownEvent) -- register for events that affect cooldowns
	Puffin.RegisterCallback("UNIT_INVENTORY_CHANGED", CooldownEvent)
	Puffin.RegisterCallback("ACTIONBAR_UPDATE_COOLDOWN", CooldownEvent)
	Puffin.RegisterCallback("BAG_UPDATE_COOLDOWN", CooldownEvent)
	Puffin.RegisterCallback("PET_BAR_UPDATE_COOLDOWN", CooldownEvent)
	Puffin.RegisterCallback("RUNE_POWER_UPDATE", CooldownEvent)
	Puffin.RegisterCallback("RUNE_TYPE_UPDATE", CooldownEvent)
	Puffin.RegisterCallback("VEHICLE_UPDATE", CooldownEvent)
	
	Puffin.RegisterCallback("UNIT_SPELLCAST_START", CheckGCD) -- register for events to track global cooldown
	Puffin.RegisterCallback("UNIT_SPELLCAST_SUCCEEDED", CheckGCD)
	
	local _
	_, _, iconGCD = GetSpellInfo(28730) -- cached for global cooldown (using same icon as Arcane Torrent)
	iconPotion = GetItemIcon(31677) -- icon for shared potions cooldown
	iconElixir = GetItemIcon(28104) -- icon for shared elixirs cooldown
end

-- Return true only if the specified runes are available, with death runes serving as wildcards
function IsRuneSpellReady(blood, frost, unholy, any)
	local b, f, u, d = 0, 0, 0, 0
	for i = 1, 6 do
		slot = runeSlots[i]
		if slot and slot.ready then
			local t = slot.rtype
			if t == 1 then b = b + 1 elseif t == 2 then u = u + 1 elseif t == 3 then f = f + 1 elseif t == 4 then d = d + 1 end
		end
	end
	if any == true then if (b + f + u + d) == 0 then return false end end -- optional test to see if any runes are available at all
	if blood and (b < 1) then if d < 1 then return false else d = d - 1 end end
	if frost and (f < 1) then if d < 1 then return false else d = d - 1 end end
	if unholy then if u < 1 and d < 1 then return false end end
	return true
end

-- Check if a spell is on cooldown because a rune is not available, return true only if on real cooldown
local function CheckRuneCooldown(name, duration)
	local runeSpells = DB.GetRecord("spells", "runes") -- get table containing rune spells
	local localSpells = DB.GetRecord("spells", "local") -- get table to map localized names
	if not runeSpells or not localSpells then return true end
	
	local spell = localSpells[name] or name -- translate to english spell name
	if spell then
		local runes = runeSpells[spell]
		if runes then
			if IsRuneSpellReady(runes.blood, runes.frost, runes.unholy, nil) then return true end -- runes are available so real cooldown
			if (duration >= 9) and (duration <= 10) then return false end -- right duration to be a rune cooldown
		end
	end
	return true
end

-- On update, manage database of active cooldowns in response to cooldown-related events
function Model_Cooldowns.OnUpdate()
	if cooldown_changed then
		local cd = cooldown
		DB.ReleaseRecords("cooldowns") -- release all cooldowns
			
		if Puffin.myClass == "DEATHKNIGHT" then -- if death knight then look for rune availability
			local blood, frost, unholy, death = 0, 0, 0, 0
			for i = 1, 6 do
				local rune, rtype = runeSlots[i], GetRuneType(i)
				local start, duration, ready = GetRuneCooldown(i)
				if not rune then
					rune = { rtype = rtype, start = start, duration = duration, ready = ready }
					runeSlots[i] = rune
				else
					rune.rtype = rtype; rune.start = start; rune.duration = duration; rune.ready = ready
				end
			end
		end

		local lockedOut = false -- flag set if any lockout spells are found
		for school in pairs(lockouts) do lockouts[school] = 0 end -- clear any previous settings in lockout table
		local lockSpells = DB.GetRecord("spells", "locks") -- get table containing lockout spells
		if lockSpells then
			for _, ls in pairs(lockSpells) do
				if not lockouts[ls.school] then lockouts[ls.school] = 0 end -- initialize when school seen for first time
				if ls.id and (lockouts[ls.school] == 0) then
					local start, duration = GetSpellCooldown(ls.id)
					if start and (start > 0) and (duration > 1.5) then -- locked out!
						lockouts[ls.school] = duration; lockstarts[ls.school] = start; lockedOut = true
						cd.cdType = "alert"; cd.id = nil; cd.icon = iconGCD; cd.start = start; cd.duration = duration; cd.count = nil
						cd.ttType = "text"; cd.ttArg = "Locked out of " .. ls.school .. " of magic."
						DB.SetRecord("cooldowns", ls.school .. " School", cd, true)
					end
				end
			end
		end

		for tab = 1, 2 do -- scan first two tabs of player spell book (general and current spec) for player spells on cooldown
			local _, _, offset, numSpells = GetSpellTabInfo(tab)
			for i = 1, numSpells do
				local index = i + offset
				local stype, id = GetSpellBookItemInfo(index, "spell")
				if stype == "SPELL" then -- use spellbook index to check for cooldown
					local start, duration, enable, count, charges
					count, charges, start, duration = GetSpellCharges(index, "spell")
					if count and charges and count < charges then enable = 1 else start, duration, enable = GetSpellCooldown(index, "spell") end
					if start and (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns
						local name, _, icon = GetSpellInfo(index, "spell")
						if name then -- make sure we have a valid spell name
							if (Puffin.myClass ~= "DEATHKNIGHT") or CheckRuneCooldown(name, duration) then -- if death knight check rune cooldown
								local locked = false
								if lockedOut then -- check if this spell is on same cooldown as any lockout spell
									for ls, ld in pairs(lockouts) do if ld == duration and lockstarts[ls] == start then locked = true end end
								end
								if not locked then
									local link = GetSpellLink(index, "spell")
									cd.cdType = "spell"; cd.id = id; cd.icon = icon; cd.start = start; cd.duration = duration; cd.count = count
									cd.ttType = "spell link"; cd.ttArg = link
									DB.SetRecord("cooldowns", name, cd, true)
								end
							end
						end
					end
				elseif stype == "FLYOUT" then -- use spell id to check for cooldown
					local _, _, numSlots = GetFlyoutInfo(id)
					for slot = 1, numSlots do
						local spellID = GetFlyoutSlotInfo(id, slot)
						if spellID then
							local start, duration, enable = GetSpellCooldown(spellID)
							if start and (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns
								local name, _, icon = GetSpellInfo(spellID)
								if name then -- make sure we have a valid spell name
									if (Puffin.myClass ~= "DEATHKNIGHT") or CheckRuneCooldown(name, duration) then -- if death knight check rune cooldown
										local locked = false
										if lockedOut then -- check if this spell is on same cooldown as any lockout spell
											for ls, ld in pairs(lockouts) do if ld == duration and lockstarts[ls] == start then locked = true end end
										end
										if not locked then
											local link = GetSpellLink(spellID)
											cd.cdType = "spell"; cd.id = spellID; cd.icon = icon; cd.start = start; cd.duration = duration; cd.count = nil
											cd.ttType = "spell link"; cd.ttArg = link
											DB.SetRecord("cooldowns", name, cd, true)
										end
									end
								end
							end
						end
					end
				end
			end
		end

		local numSpells = HasPetSpells() -- scan pet spell book for spells on cooldown
		if numSpells and UnitExists("pet") then
			for index = 1, numSpells do
				local start, duration, enable = GetSpellCooldown(index, "pet")
				if start and (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns
					local name, _, icon = GetSpellInfo(index, "pet") -- this info could all be cached and updated with spell book changes
					if name then
						local link = GetSpellLink(index, "pet")
						local _, id = GetSpellBookItemInfo(index, "pet")
						cd.cdType = "pet"; cd.id = id; cd.icon = icon; cd.start = start; cd.duration = duration; cd.count = nil
						cd.ttType = "spell link"; cd.ttArg = link
						DB.SetRecord("cooldowns", name, cd, true)
					end
				end
			end
		end

		local offset = nil -- check for override/vehicle bar actions on cooldown
		if HasVehicleActionBar() then offset = 132 elseif HasOverrideActionBar() then offset = 156 end
		if offset then
			for slot = 1, 6 do
				local actionType, id = GetActionInfo(slot + offset)
				if actionType == "spell" then
					local start, duration, enable = GetSpellCooldown(id)
					if start and (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns
						local name, _, icon = GetSpellInfo(id)
						if name then
							cd.cdType = "possess"; cd.id = id; cd.icon = icon; cd.start = start; cd.duration = duration; cd.count = nil
							cd.ttType = "spell id"; cd.ttArg = id
							DB.SetRecord("cooldowns", name, cd, true)
						end
					end
				end
			end
		end

		for bag = 0, NUM_BAG_SLOTS do -- check in bags for item cooldowns
			local numSlots = GetContainerNumSlots(bag)
			for i = 1, numSlots do
				local itemID = GetContainerItemID(bag, i)
				if itemID then
					local start, duration, enable = GetItemCooldown(itemID)
					if start and (start > 0) and (enable == 1) and (duration > 1.5) then -- don't include global cooldowns or really short cooldowns
						local name, link, _, _, _, itemType, itemSubType, _, _, icon = GetItemInfo(itemID)
						if name then
							local found = false
							if itemType == "Consumable" then -- check for shared cooldowns for potions/elixirs/flasks
								if itemSubType == "Potion" then
									found = true
									if not DB.GetRecord("cooldowns", "Potions") then
										cd.cdType = "item"; cd.id = nil; cd.icon = iconPotion; cd.start = start; cd.duration = duration; cd.count = nil
										cd.ttType = "text"; cd.ttArg = "Shared Potion Cooldown"
										DB.SetRecord("cooldowns", "Potions", cd, true)
									end
								elseif (itemSubType == "Elixir") or (itemSubType == "Flask") then
									found = true
									if not DB.GetRecord("cooldowns", "Elixirs") then
										cd.cdType = "item"; cd.id = nil; cd.icon = iconElixir; cd.start = start; cd.duration = duration; cd.count = nil
										cd.ttType = "text"; cd.ttArg = "Shared Elixir Cooldown"
										DB.SetRecord("cooldowns", "Elixirs", cd, true)
									end
								end
							end
							if not found then
								cd.cdType = "item"; cd.id = nil; cd.icon = icon; cd.start = start; cd.duration = duration; cd.count = nil
								cd.ttType = "item link"; cd.ttArg = link
								DB.SetRecord("cooldowns", name, cd, true)
							end
						end
					end
				end
			end
		end

		for _, slot in pairs(slots) do -- get cooldown info for all inventory slots
			local id = GetInventorySlotInfo(slot)
			if id then
				local start, duration, enable = GetInventoryItemCooldown("player", id)
				if start and (start > 0) and (enable == 1) and (duration > 1.5) then
					local name, _, _, _, _, _, _, _, _, icon = GetItemInfo(GetInventoryItemLink("player", id))
					cd.cdType = "item"; cd.id = nil; cd.icon = icon; cd.start = start; cd.duration = duration; cd.count = nil
					cd.ttType = "inventory"; cd.ttArg = slot
					DB.SetRecord("cooldowns", name, cd, true)
				end
			end
		end
		
		if startGCD and durationGCD then -- detect global cooldowns
			if (startGCD + durationGCD) > GetTime() then
				cd.cdType = "GCD"; cd.id = nil; cd.icon = iconGCD; cd.start = startGCD; cd.duration = durationGCD; cd.count = nil
				cd.ttType = "text"; cd.ttArg = "GCD"
				DB.SetRecord("cooldowns", "GCD", cd, true)
			else
				startGCD = nil; durationGCD = nil -- this cooldown is no longer valid
			end
		end
	end
	
	for name, cd in DB.GetRecords("cooldowns") do
		local t = cd.start + cd.duration - GetTime() -- update cooldowns with timeLeft and release expired cooldowns
		if t > 0 then cd.timeLeft = t else DB.ReleaseRecord("cooldowns", name); cooldown_changed = true end
	end
--[[	
	if IsAltKeyDown() then	
		for name, cd in DB.GetRecords("cooldowns") do
			print("cd", name, cd.id, cd.timeLeft, cd.duration, cd.count or "")
		end
	end
]]--
	if cooldown_changed then cooldown_changed = false; Puffin.Callback("Cooldowns") end
end

Puffin.RegisterComponent("Model_Cooldowns", Model_Cooldowns, "MODEL", 4)
