-- Puffin: A Tojaso FowlPlay Project
-- View component for a data broker-compatible plugin to display with a single icon and single text field.

-- View API:
-- View_BrokerBlock.Create() - a new instance of a broker block view
-- View_BrokerBlock:Release() - release an instance of a broker block view, recycling it for future use
-- View_BrokerBlock:Configure() - configure the broker block based on current defaults
-- View_BrokerBlock:Update() - update the broker block

-- View attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- button - widget template for button [c]
-- text - widget template for text field [c]
-- offsetX - horizontal offset from UIParent [c]
-- offsetY - vertical offset from UIParent [c]
-- anchorPoint - if anchor defined then point on UIParent to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [c]
-- alpha - alpha for the broker block [c]
-- hasIcon - true if the broker should show the icon
-- hasText - true if the broker block should show the text field
-- hidden - hide the view [d]

local View_BrokerBlock = {}
local EF = Puffin.ErrorFormat
local DB, LSM
local recycledViews = {}
local viewNumber = 0
local classDepth = 0 -- mininum depth for view's of this class

-- Return an initialized instance of the view
function View_BrokerBlock.Create()
	local n, view = next(recycledViews) -- get any available recycled widget
	if n then
		recycledViews[n] = nil
	else
		view = {}
		viewNumber = viewNumber + 1
		view._widgets = {}
	end
	return view
end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_BrokerBlock:Release()
	self._buttonWidget._class.Release(self._buttonWidget); self._buttonWidget = nil -- release the widgets used by this view
	self._textWidget._class.Release(self._textWidget); self._textWidget = nil
	for k in pairs(self) do if (k ~= "_buttonWidget") and (k ~= "_textWidget") then self[k] = nil end end -- clear attributes for this view
	table.insert(recycledViews, self)
end

-- Check an attribute to see if it refers to special media
local function CheckMedia(v)
	if type(v) == "table" and v[1] == "#LSM" then v = LSM:Fetch(v[2], v[3]) end -- check for LibSharedMedia references
	return v
end

-- Handle mouse interactions with the broker block view
local function BrokerBlock_OnEnter(frame, broker, ...)
	if broker then
		-- print("onEnter", broker.label or broker.text or "unknown", broker.tooltip, broker.OnEnter, broker.OnTooltipShow)
		if broker.tooltip then
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT")
			if broker.tooltiptext then broker.tooltip:SetText(broker.tooltiptext) end
			broker.tooltip:Show()
		elseif broker.OnEnter then
			broker.OnEnter(frame, ...)
			GameTooltip:ClearAllPoints()
			GameTooltip:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT")
			GameTooltip:Show()
		elseif broker.OnTooltipShow then
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT")
			broker.OnTooltipShow(GameTooltip)
			GameTooltip:Show()
		else
			local text = broker.text or broker.label
			if broker.tooltiptext then text = broker.tooltiptext end
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT")
			GameTooltip:SetText(text or "Unknown")
			GameTooltip:Show()
		end
	end
end

local function BrokerBlock_OnLeave(frame, broker, ...)
	if broker then
		-- print("onLeave", broker.label or broker.text or "unknown")
		if broker.tooltip then
			GameTooltip:Hide()
		elseif broker.OnLeave then
			broker.OnLeave(frame, ...)
		else
			GameTooltip:Hide()
		end
	end
end

local function BrokerBlock_OnClick(frame, broker, ...)
	if broker and broker.OnClick then broker.OnClick(frame, ...) end
end

-- Configure a widget based on its descriptor
local function ConfigureWidget(bb, t, w, offsetDepth)
	t._class.SetPosition(t, w.anchorPoint or bb.anchorPoint or "BOTTOMLEFT", w.anchor or nil, w.relativePoint or bb.relativePoint or "BOTTOMLEFT",
		w.offsetX or 0, w.offsetY or 0)
	t.width = (w.width or 0) + ((w.relativeWidth or 0) * bb.width); t.height = (w.height or 0) + ((w.relativeHeight or 0) * bb.height)
	t.iconSize = bb.iconSize or 0; t.depth = (w.depth or 0) + offsetDepth
	t.scale = w.scale or 1; t.alpha = w.alpha or 1
	if w.attributes then for i, v in pairs(w.attributes) do t[i] = CheckMedia(v) end end -- widget specific attributes
	local defaults = DB.GetRecord("widget_defaults", bb.defaults)
	if defaults and defaults[w.class] then -- default attributes for this widget class
		for i, v in pairs(defaults[w.class]) do if t[i] == nil then t[i] = CheckMedia(v) end end
	end
	t._class.Configure(t)
end

-- Create widget based on its descriptor
local function CreateWidget(bb, w)
	local class = Puffin.Component(w.class)
	assert(class, EF("View_BrokerBlock", "unrecognized class"))
	t = class.Create(); t._class = class
	return t
end

-- Configure all widgets associated with this view
function View_BrokerBlock:Configure()
	assert(self.defaults, EF("View_BrokerBlock", "undefined defaults"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10
	
	if not self._buttonWidget and self.button then
		self.button.offsetX = self.offsetX; self.button.offsetY = self.offsetY
		local t = CreateWidget(self, self.button)
		if t then
			self._buttonWidget = t
			t.onEnter = BrokerBlock_OnEnter; t.onLeave = BrokerBlock_OnLeave; t.onClick = BrokerBlock_OnClick; t.masque = "Broker"
			ConfigureWidget(self, t, self.button, offsetDepth)
		end
	end

	if not self._textWidget and self.text and self.hasText then
		self.text.width = self.textWidth; self.text.offsetX = self.hasIcon and (self.iconSize + 3) or 0; self.text.offsetY = 0
		self.text.anchor = self._buttonWidget; self.text.anchorPoint = "BOTTOMLEFT"; self.text.relativePoint = "BOTTOMLEFT"
		local t = CreateWidget(self, self.text)
		if t then
			self._textWidget = t
			ConfigureWidget(self, t, self.text, offsetDepth)
		end
	end	
end

-- Update all widgets associated with this view, using widget-specific providers to access the data record
function View_BrokerBlock:Update(broker)
	local hide = self.hidden
	local t = self._textWidget
	if t then
		t.hidden = hide
		if not hide and self.text and self.hasText then t.text = broker.text else t.text = nil end
		t._class.Update(t)
	end
	t = self._buttonWidget
	if t then
		t.hidden = hide
		if not hide and self.button and self.hasIcon then t.texture = broker.icon else t.texture = nil end
		t.id = broker; t._class.Update(t)
	end
end

function View_BrokerBlock.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_BrokerBlock", "Data component not found"))
	LSM = LibStub("LibSharedMedia-3.0")
	assert(LSM, EF("View_BrokerBlock", "LibSharedMedia not found"))
end

Puffin.RegisterComponent("View_BrokerBlock", View_BrokerBlock, "VIEW", 0)
