-- Puffin: A Tojaso FowlPlay Project
-- View component for an icon group that shows icons for all spells in a list.
-- Each icon is assembled from a collection of widgets, as defined by a layout.

-- Defaults: default attribute values for each class of widget used by the view
-- Widgets: templates that specify exactly which widgets instances go into the view

-- Widget templates include the following attributes:
-- class - widget class name [c]
-- priority - if true then configure this widget before others [c]
-- parent - parent widget for scale and alpha adjustments [c]
-- anchor - widget to anchor to (default: anchor to UIParent plus view's offsetX and offsetY) [c]
-- anchorPoint - if anchor defined then point on anchor to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- offsetX - horizontal offset from anchor (default: 0) [c]
-- offsetY - vertical offset from anchor (default: 0) [c]
-- scale - factor to scale widget's display size [c]
-- depth - relative to view's depth (range: 0 .. 9) [c]
-- alpha - opacity for the widget [d]
-- hidden - hide the widget [d]

-- View API:
-- View_IconGroup.Create() - a new instance of an icon group view
-- View_IconGroup:Release() - release an instance of an icon group view, recycling it for future use
-- View_IconGroup:Configure() - configure based on current defaults and widget descriptions
-- View_IconGroup:Update() - update the icon group

-- View attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- widgets - table of widget templates actually used to construct this view [c]
-- offsetX - horizontal offset from BOTTOMLEFT of overall screen to BOTTOMLEFT of interactive area (default: 0) [c]
-- offsetY - vertical offset from BOTTOMLEFT of overall screen to BOTTOMLEFT of interactive area (default: 0) [c]
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [c]
-- alpha - opacity for the view [d]
-- hidden - hide the view [d]

local View_IconGroup = {}
local EF = Puffin.ErrorFormat
local DB, LSM
local recycledViews = {}
local viewNumber = 0
local classDepth = 0 -- mininum depth for view's of this class

-- Return an initialized instance of the view
function View_IconGroup.Create()
	local n, view = next(recycledViews) -- get any available recycled widget
	if n then
		recycledViews[n] = nil
	else
		view = {}
		viewNumber = viewNumber + 1
		view._widgets = {}
	end
	return view
end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_IconGroup:Release()
	for k, w in pairs(self._widgets) do w._class.Release(w); self._widgets[k] = nil end -- release the other widgets used by this view
	for k in pairs(self) do if (k ~= "_widgets") then self[k] = nil end end -- clear attributes for this view
	table.insert(recycledViews, self)
end

-- Check an attribute to see if it refers to special media
local function CheckMedia(v)
	if type(v) == "table" and v[1] == "#LSM" then v = LSM:Fetch(v[2], v[3]) end -- check for LibSharedMedia references
	return v
end

-- Configure a widget based on its descriptor
local function ConfigureWidget(dp, k, w, offsetDepth)
	local t = dp._widgets[k] -- check if already allocated this widget
	if not t then -- need to create an instance of the widget
		local class = Puffin.Component(w.class)
		assert(class, EF("View_IconGroup", "unrecognized class"))
		t = class.Create(); t._class = class; dp._widgets[k] = t
	end
	local x, y, a = (w.offsetX or 0) + ((w.relativeX or 0) * dp.width), (w.offsetY or 0) + ((w.relativeY or 0) * dp.height), w.anchor
	if a then a = dp._widgets[a] or nil else x = x + (dp.offsetX or 0); y = y + (dp.offsetY or 0) end
	t._class.SetPosition(t, w.anchorPoint or dp.anchorPoint or "BOTTOMLEFT", a, w.relativePoint or dp.relativePoint or "BOTTOMLEFT", x, y, w.anchorRegion)
	if w.parent then t._class.SetParent(t, dp._widgets[w.parent]) end
	t.width = (w.width or 0) + ((w.relativeWidth or 0) * dp.width); t.height = (w.height or 0) + ((w.relativeHeight or 0) * dp.height)
	t.depth = (w.depth or 0) + offsetDepth
	t.scale = w.scale or 1; t.alpha = w.alpha or 1
	if w.attributes then for i, v in pairs(w.attributes) do t[i] = CheckMedia(v) end end -- widget specific attributes
	local defaults = DB.GetRecord("widget_defaults", dp.defaults)
	if defaults and defaults[w.class] then -- default attributes for this widget class
		for i, v in pairs(defaults[w.class]) do if t[i] == nil then t[i] = CheckMedia(v) end end
	end
	t._class.Configure(t)
end

-- Configure all widgets associated with this view
function View_IconGroup:Configure()
	assert(self.defaults, EF("View_IconGroup", "undefined defaults"))
	assert(self.widgets, EF("View_IconGroup", "undefined widgets"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10
end

-- Update all widgets associated with this view, using widget-specific providers to access the data record
function View_IconGroup:Update(record, unit)
	local templates = DB.GetRecord("widget_templates", self.templates)
	local widgets = DB.GetRecord("widgets", self.widgets)
	if widgets and templates then
		for _, w in pairs(widgets) do -- each widget has unique name as key into templates table
			local template = templates[w]
			local t = self._widgets[w]
			assert(t, EF("View_IconGroup", "update called for invalid widget"))
			t.hidden = self.hidden
			if not t.hidden and template.providers then -- call providers for widget specific data attributes
				for i, p in pairs(template.providers) do
					if record and type(p) == "table" then
						local v = DB.GetAttribute("providers", p.class, p.name)
						if type(v) == "function" and not p.no_function then v = v(record, unit, p.param) end
						if type(v) == "table" and not p.no_table and v[i] then for j, k in pairs(v) do t[j] = k end else t[i] = v end
					elseif i == "condition" then  -- special case to test for existence of the unit
						t[i] = (p == "exists") and (record ~= nil) or false
					end
				end
			end
			t._class.Update(t)
		end
	end
end

function View_IconGroup.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_IconGroup", "Data component not found"))
	LSM = LibStub("LibSharedMedia-3.0")
	assert(LSM, EF("View_IconGroup", "LibSharedMedia not found"))
end

Puffin.RegisterComponent("View_IconGroup", View_IconGroup, "VIEW", 0)
